@@ -195,7 +195,6 @@ class UFOOperator(object):
         self.mutedAxisNames = None    # list of axisname that need to be muted
         self.strict = strict
         self.debug = debug
-        self.logger = None
         self.extrapolate = extrapolate  # if true allow extrapolation
         self.logger = None
         self.doc = None
@@ -207,17 +206,22 @@ class UFOOperator(object):
         else:
             self.doc = DesignSpaceDocument()
         if self.debug:
-            docBaseName = os.path.splitext(self.doc.path)[0]
-            logPath = f"{docBaseName}_log.txt"
-            self.logger = Logger(path=logPath, rootDirectory=None)
-            self.logger.time()
-            self.logger.info(f"## {self.doc.path}")
-            self.logger.info(f"\tUFO version: {self.ufoVersion}")
-            self.logger.info(f"\tround Geometry: {self.roundGeometry}")
-            if self.useVarlib:
-                self.logger.info(f"\tinterpolating with varlib")
-            else:
-                self.logger.info(f"\tinterpolating with mutatorMath")
+            self.startLog()
+
+    def startLog(self):
+        # so we can call it later
+        self.debug = True
+        docBaseName = os.path.splitext(self.doc.path)[0]
+        logPath = f"{docBaseName}_log.txt"
+        self.logger = Logger(path=logPath, rootDirectory=None)
+        self.logger.time()
+        self.logger.info(f"## {self.doc.path}")
+        self.logger.info(f"\tUFO version: {self.ufoVersion}")
+        self.logger.info(f"\tround Geometry: {self.roundGeometry}")
+        if self.useVarlib:
+            self.logger.info(f"\tinterpolating with varlib")
+        else:
+            self.logger.info(f"\tinterpolating with mutatorMath")
 
     def _instantiateFont(self, path):
         """ Return a instance of a font object with all the given subclasses"""
@@ -261,6 +265,12 @@ class UFOOperator(object):
     def addAxisDescriptor(self, **kwargs):
         return self.doc.addAxisDescriptor(**kwargs)
 
+    def addLocationLabel(self, locationLabelDescriptor):
+        self.doc.addLocationLabel(locationLabelDescriptor)
+
+    def addLocationLabelDescriptor(self, **kwargs):
+        return self.doc.addLocationLabelDescriptor(**kwargs)
+
     def addRule(self, ruleDescriptor):
         self.doc.addRule(ruleDescriptor)
 
@@ -389,11 +399,11 @@ class UFOOperator(object):
     # loading and updating fonts
     def loadFonts(self, reload=False):
         # Load the fonts and find the default candidate based on the info flag
+        self.glyphNames = list({glyphname for font in self.fonts.values() for glyphname in font.keys()})
         if self._fontsLoaded and not reload:
             if self.debug:
-                self.logger.info("\t\t-- loadFonts requested, but fonts are loaded already and no reload requested")
+                self.logger.info("\t\t-- loadFonts called, but fonts are loaded already and no reload requested")
             return
-        names = set()
         actions = []
         if self.debug:
             self.logger.info("## loadFonts")
@@ -407,12 +417,10 @@ class UFOOperator(object):
                     thisLayerName = getDefaultLayerName(font)
                     if self.debug:
                         actions.append(f"loaded: {os.path.basename(sourceDescriptor.path)}, layer: {thisLayerName}, format: {font.ufoFormatVersionTuple}, id: {id(font):X}")
-                    names |= set(self.fonts[sourceDescriptor.name].keys())
                 else:
                     self.fonts[sourceDescriptor.name] = None
                     if self.debug:
                         actions.append("source ufo not found at %s" % (sourceDescriptor.path))
-        self.glyphNames = list(names)
         if self.debug:
             for item in actions:
                 self.logger.infoItem(item)
@@ -591,6 +599,7 @@ class UFOOperator(object):
 
     def splitLocation(self, location):
         # split a location in a continouous and a discrete part
+        # Note: discrete can be None
         discreteAxes = [a.name for a in self.getOrderedDiscreteAxes()]
         continuous = {}
         discrete = {}
@@ -1268,6 +1277,17 @@ class UFOOperator(object):
             locHorizontal, locVertical = self.splitAnisotropic(loc)
         return anisotropic, continuousLocation, discreteLocation, locHorizontal, locVertical
 
+    def collectSkippedGlyphs(self):
+        # return a list of all the glyphnames listed in public.skipExportGlyphs
+        names = []
+        for fontPath, fontObj in self.fonts.items():
+            for name in fontObj.lib.get('public.skipExportGlyphs', []):
+                if name not in names:
+                    names.append(name)
+        if self.debug:
+            self.logger.info(f"collectSkippedGlyphs: {names}")
+        return names
+
     def makeInstance(self, instanceDescriptor,
             doRules=None,
             glyphNames=None,
@@ -1282,6 +1302,7 @@ class UFOOperator(object):
         # hmm getFullDesignLocation does not support anisotropc locations?
         fullDesignLocation = instanceDescriptor.getFullDesignLocation(self.doc)
         anisotropic, continuousLocation, discreteLocation, locHorizontal, locVertical = self.getLocationType(fullDesignLocation)
+        self.loadFonts()
 
         if not self.extrapolate:
            # Axis values are in userspace, so this needs to happen before bending
@@ -1296,6 +1317,7 @@ class UFOOperator(object):
             locHorizontal, locVertical = self.splitAnisotropic(loc)
             if self.debug:
                 self.logger.info(f"\t\t\tAnisotropic location for \"{instanceDescriptor.name}\"\n\t\t\t{fullDesignLocation}")
+
         # makeOneKerning
         # discreteLocation ?
         if instanceDescriptor.kerning:
@@ -1323,7 +1345,6 @@ class UFOOperator(object):
             # each key in the location is the libKey
             # each value is the calculated value
             libMathDict = libMathMutator.makeInstance(locHorizontal)
-            #print("libMathDict", locHorizontal, libMathDict)
             if libMathDict:
                 for libKey, mutatedValue in libMathDict.items():
                     # only add the value to the lib if it is not 0.
@@ -1361,7 +1382,9 @@ class UFOOperator(object):
         if 'public.glyphOrder' not in font.lib.keys():
             # should be the glyphorder from the default, yes?
             font.lib['public.glyphOrder'] = selectedGlyphNames
-
+        # remove skippable glyphs
+        toSkip = self.collectSkippedGlyphs()
+        selectedGlyphNames = [name for name in selectedGlyphNames if name not in toSkip]
         for glyphName in selectedGlyphNames:
             glyphMutator, unicodes = self.getGlyphMutator(glyphName, decomposeComponents=decomposeComponents, discreteLocation=discreteLocation)
             if glyphMutator is None:
@@ -1427,14 +1450,22 @@ class UFOOperator(object):
 
     def locationToDescriptiveString(self, loc):
         # make a nice descriptive string from the location
+        # Check if the discrete location is None.
         t = []
         cl, dl = self.splitLocation(loc)
         for continuousAxis in sorted(cl.keys()):
             t.append(f'{continuousAxis}_{cl[continuousAxis]}')
-        for discreteAxis in sorted(dl.keys()):
-            t.append(f'{discreteAxis}_{dl[discreteAxis]}')
+        if dl is not None:
+            for discreteAxis in sorted(dl.keys()):
+                t.append(f'{discreteAxis}_{dl[discreteAxis]}')
         return '_'.join(t)
 
+    def pathForInstance(self, instanceDescriptor):
+        # generate the complete path for this instance descriptor.
+        if self.path is not None and instanceDescriptor.filename is not None:
+            return os.path.abspath(os.path.join(os.path.dirname(self.path), instanceDescriptor.filename))
+        return None
+
     def makeOneInstance(self, location,
             doRules=None,
             glyphNames=None,
@@ -1444,6 +1475,7 @@ class UFOOperator(object):
         # make one instance for this location. This is a shortcut for making an
         # instanceDescriptor. So it makes some assumptions about the font names.
         # Otherwise all the geometry will be exactly what it needs to be.
+        self.loadFonts()
         continuousLocation, discreteLocation = self.splitLocation(location)
         defaultFont = self.findDefaultFont(discreteLocation=discreteLocation)
         if defaultFont is not None:
@@ -1520,6 +1552,7 @@ class UFOOperator(object):
     # @memoize
     def makeFontProportions(self, location, bend=False, roundGeometry=True):
         """Calculate the basic font proportions for this location, to map out expectations for drawing"""
+        self.loadFonts()
         continuousLocation, discreteLocation = self.splitLocation(location)
         infoMutator = self.getInfoMutator(discreteLocation=discreteLocation)
         data = dict(unitsPerEm=1000, ascender=750, descender=-250, xHeight=500)
@@ -1553,8 +1586,8 @@ class UFOOperator(object):
 
         Returns: a mathglyph, results are cached
         """
+        self.loadFonts()
         continuousLocation, discreteLocation = self.splitLocation(location)
-
         bend=False  #
         if not self.extrapolate:
             # Axis values are in userspace, so this needs to happen *after* clipping.
@@ -1605,6 +1638,7 @@ class UFOOperator(object):
         """
         if self.debug:
             self.logger.info(f"\t\t\tmakeOneInfo for {location}")
+        self.loadFonts()
         bend = False
         anisotropic, continuousLocation, discreteLocation, locHorizontal, locVertical = self.getLocationType(location)
         # so we can take the math object that comes out of the calculation
@@ -1635,6 +1669,7 @@ class UFOOperator(object):
         """
         if self.debug:
             self.logger.info(f"\t\t\tmakeOneKerning for {location}")
+        self.loadFonts()
         bend = False
         kerningObject = None
         anisotropic, continuousLocation, discreteLocation, locHorizontal, locVertical = self.getLocationType(location)
@@ -1653,6 +1688,8 @@ class UFOOperator(object):
                 # extract the object later
                 if self.debug:
                     self.logger.info(f"\t\t\t\t{len(kerningObject.keys())} kerning pairs added")
+        if self.roundGeometry:
+            kerningObject.round()
         if self.debug:
             if kerningObject is not None:
                 self.logger.info(f"\t\t\t\tmakeOneKerning outcome: {kerningObject.items()}")
@@ -1733,7 +1770,6 @@ if __name__ == "__main__":
     doc.libKeysForProcessing.append(testLibMathKey)
     print('processing these keys', doc.libKeysForProcessing)
 
-
     if makeUFOs:
         doc.generateUFOs()
     randomLocation = doc.randomLocation()
@@ -1845,3 +1881,8 @@ if __name__ == "__main__":
             print('italicslantoffset at', randomLocation, m.makeInstance(randomLocation))
         else:
             print("getLibEntryMutator() returned None.")
+
+    for instanceDescriptor in doc.instances:
+        print('path for instancedescriptor', doc.pathForInstance(instanceDescriptor))
+
+    doc.collectSkippedGlyphs()
\ No newline at end of file
